/*******************************************************************

	Stream Binary File Transfer Class Implementation

	(c) Copyright 1992/1993 Marco Savegnago IW3FQG

*******************************************************************/

#include <xferprot/tstream.h>
#include <dos/ttimeout.h>

#define __MISC_INLINE   // usa routine inline
#include <dos/tmisc.h>

#if 0
#define DEBUGTSTREAM
#else
#undef  DEBUGTSTREAM
#endif

#ifdef  DEBUGTSTREAM
#include "mstream.h"
#endif

TStreamXfer::TStreamXfer(TStream *port, fstream* afstream, xferType type) :
	TBinaryXferProtcol(port, afstream, type)
{
	theCurrentBlock=0;
	lastError = None;
}

Boolean TStreamXfer::xfer()
{
	TStreamMode old = theStream->setMode(moUnPolled|moBinary);

	Boolean result = (theXferType) ? transmit() : receive();

	theStream->setMode(old);

	return result;
}

Boolean TStreamXfer::waitReady()
{
	if (theStream->isAny())
					return False;   // Se c'e' qualcosa in RX torna subito

	TTimeout timeout(2184);	// 120 secondi

	while(!timeout && theStream->isConnected())
		if (theStream->isAny())
            		return False;

    lastError=Timeout;

    return True;
}

Boolean TStreamXfer::sendLenOfFile()
{
    struct TStartHdr : public TStreamBlockHead
    {
            byte    type;
            byte    lenght;
            dword   value;
            word    crc;

    } aBlock;

    aBlock.TInfoBlock.type=FileInfo;
    aBlock.TInfoBlock.lenght=sizeof(TStartHdr) - sizeof(TStreamBlockHead) - 2;
    aBlock.TInfoBlock.number=theCurrentBlock=0;

    aBlock.type=LENOFFILE;
    aBlock.lenght=4;                // len of Filesize

    fileSize=fstreamLenght(theFileStream);

    aBlock.value=networkize(fileSize);

    TCRC16 crc((byte *) &aBlock, sizeof(aBlock) - 2); // All but CRC

    aBlock.crc=networkize(word(crc));

    theStream->put((byte*)&aBlock, sizeof(aBlock));

    theCurrentBlock++;

    return True;
}

Boolean TStreamXfer::recvBlock(word blockLen)
{
    word    len=0;
    byte frameBuf[1024];

    word onechunk = sizeof frameBuf;

    while((len + onechunk) <= blockLen)
    {
        if(waitReady())
            	return False;

	if(!theStream->get(frameBuf, onechunk))
	{
            lastError = ReadingPort;
            return False;
        }

        theCurrentBlockCRC.updateCRCBlock(frameBuf, onechunk);

        if(!write(frameBuf, onechunk))
		{
        	lastError=WritingFile;
        	return False;
        }

        len += onechunk;

        xferCount += len;
    }

    if(waitReady())
        return False;

    word rest = blockLen - len;

    word blockCrc;

    if(rest)
    {
	if(!theStream->get(frameBuf, rest + 2)) // include CRC
    	{
        	lastError = ReadingPort;
        	return False;
    	}

    	theCurrentBlockCRC.updateCRCBlock(frameBuf, rest);

    	blockCrc = intelize(word(*(frameBuf+rest) | (*(frameBuf+rest+1) << 8)));
    }
    else
    {
	if(!theStream->get(frameBuf, 2)) // include CRC
    	{
        	lastError = ReadingPort;
        	return False;
	}

    	blockCrc = intelize(word(*(frameBuf) | ( *(frameBuf+1) << 8)) );
    }

    if( theCurrentBlockCRC != blockCrc)
    {
#ifdef DEBUGTSTREAM
	if(theStream->rxOverRun())
		mout << "RX OverRun!!!! " << endl;
#endif
       	lastError = CrcError;
       	return False;
    }

    if(!write(frameBuf, rest))
    {
    	lastError=WritingFile;
        return False;
    }

    xferCount += rest;

#ifdef DEBUGTSTREAM
    mout << "Stream Block " << (int)theCurrentBlock << " FileSize " << xferCount << endl;
#endif

    return True;
}

Boolean TStreamXfer::getFileInfo(TStreamBlockHead& aBlock)
{
    struct TStartHdr
    {
        byte    type;
		byte    lenght;
        dword   value;
        word    crc;

    } anotherBlock;

    if(aBlock.TInfoBlock.lenght != (sizeof anotherBlock - 2))
    {
        lastError = Abort;
        return False;
    }

    if(!theStream->get((byte *) &anotherBlock, sizeof anotherBlock))
    {
        lastError = ReadingPort;
        return False;
    }

    fileSize=intelize(anotherBlock.value);          // Filesize

    if(0!=(theCurrentBlock=aBlock.TInfoBlock.number))
    {
    	lastError = BlockError;
		*theStream << "Stream error: waiting block 0 getting " << (int) theCurrentBlock << "\r";

#ifdef DEBUGTSTREAM
		mout	<< "Stream error: waiting block 0 getting " << (int) theCurrentBlock << endl;
#endif
 		return False;
	}

#ifdef DEBUGTSTREAM
    mout << "Stream Filesize " << fileSize << endl;
#endif

    TCRC16 crc16((byte *)&aBlock, sizeof(aBlock));

    crc16.updateCRCBlock((byte *)&anotherBlock, sizeof (anotherBlock) - 2);

    if(intelize(anotherBlock.crc) != crc16)
    {
        lastError = CrcError;
		*theStream << "Stream error: CRC in block " << (int) theCurrentBlock << "\r";

#ifdef DEBUGTSTREAM
	mout << "Stream error: CRC in block " << (int) theCurrentBlock << endl;
#endif
        return False;
    }

    return True;
}


Boolean TStreamXfer::decodeInfo(TStreamBlockHead& aBlock)
{
	switch(aBlock.TInfoBlock.type)
    {
    	case fileInfoType:
        	return getFileInfo(aBlock);

        default:
        	return False;
	}
}

Boolean TStreamXfer::recvData()
{
        TStreamBlockHead aBlock;
#ifdef DEBUGTSTREAM
		mout << "Now in recvData" << endl;
#endif
        while(True)
        {
       		if(waitReady())
                {

#ifdef DEBUGTSTREAM
                    mout << "Abort because timeout" << endl;
#endif
                    break;
                }

                // Get the block header

				if(!theStream->get((byte*) &aBlock, sizeof aBlock))
                {
#ifdef DEBUGTSTREAM
					mout << "Abort because dont get the header" << endl;
#endif
        			lastError = ReadingPort;

                    break;
				}

                if(aBlock.blockType >= firstInfoType)
                {
                	if(!decodeInfo(aBlock))
                    					break;
                }
                else
                {
                	word    blockLenght=intelize(aBlock.TNormalBlock.lenght);

	                if(aBlock.TNormalBlock.number != theCurrentBlock+1)
					{
						*theStream << "Stream error: waiting block " << (int) (theCurrentBlock + 1) << " getting " << (int) aBlock.TNormalBlock.number << "\r";
#ifdef DEBUGTSTREAM
						mout << "Stream error: waiting block " << (int) theCurrentBlock+1 << " getting " << (int) aBlock.TNormalBlock.number << endl;
#endif
                    	lastError=BlockError;

                        return False;
                    }

                    theCurrentBlock=aBlock.TNormalBlock.number;

#ifdef DEBUGTSTREAM
					mout << "BlockLen = " << blockLenght << " CurrentBlockNow = " << (int) theCurrentBlock << endl;
#endif
                	if(!blockLenght)        // EOF
                	{
#ifdef DEBUGTSTREAM
						mout << "END OF TRANSFER " << endl;
#endif
						return True;
                	}

                	theCurrentBlockCRC = TCRC16((byte*)&aBlock, sizeof aBlock);

                	if(!recvBlock(blockLenght))
                    				break;
				}

        }

    	return False;
}

Boolean TStreamXfer::sendData()
{
    struct TNormalBlock : public TStreamBlockHead
    {
            byte    data[1024+2];
    } aBlock;

    while(True)
    {
        aBlock.TNormalBlock.number=theCurrentBlock++;

		if(theStream->isAny())      // Abort??
        	            break;

        word blockLen=1024;

		if(!read(aBlock.data, blockLen))
        {
        	lastError=ReadingFile;
        	break;
        }

        aBlock.TNormalBlock.lenght=networkize(blockLen);

        TCRC16 blockCrc((byte*)&aBlock, sizeof(TStreamBlockHead) + blockLen);

        word crc=networkize(blockCrc);

        *(aBlock.data+blockLen)   =   lowByte(crc);
        *(aBlock.data+blockLen+1) = highByte(crc);

		if(!theStream->put((byte*)&aBlock, sizeof(TStreamBlockHead) + blockLen + 2))
    	{
    		lastError = WritingPort;
            break;
    	}

    	xferCount += blockLen;

        if (theFileStream->eof())
        {
			TStreamBlockHead aBlock;

    		aBlock.TNormalBlock.lenght=0;
    		aBlock.TNormalBlock.number=theCurrentBlock++;

			theStream->put((byte*)&aBlock, sizeof(aBlock));

			return True;
        }

    }

    return False; // Dummy
}


Boolean TStreamXfer::receive()
{
    isRunning=True;

	recvData();

    isRunning=False;

    return xferOK = Boolean (xferCount == fileSize);
}

Boolean TStreamXfer::transmit()
{
	isRunning=True;

    if(!sendLenOfFile())
            return False;

    if(!sendData())
            return False;

    isRunning=False;

    return xferOK = Boolean (xferCount == fileSize);
}
